<?php
/**
 * @version     1.1.1
 * @package     sellacious
 *
 * @copyright   Copyright (C) 2016. Bhartiy Web Technologies. All rights reserved.
 * @license     GNU General Public License version 2 or later; see LICENSE.txt
 * @author      Izhar Aazmi <info@bhartiy.com> - http://www.bhartiy.com
 */
// no direct access.
defined('_JEXEC') or die;

use Joomla\Utilities\ArrayHelper;

// require the PayPal SDK Auto-loader
JLoader::import('paypal.autoload');
JLoader::import('sellacious.loader');

// use PayPal\Api\Amount;
// use PayPal\Api\CreditCard;
// use PayPal\Api\CreditCardToken;
// use PayPal\Api\FundingInstrument;
// use PayPal\Api\Payer;
// use PayPal\Api\Payment;
// use PayPal\Api\PaymentExecution;
// use PayPal\Api\RedirectUrls;
// use PayPal\Api\Transaction;

/**
 * Plugin to manage payment via 'PayPal' for sellacious shops checkout process
 *
 * @subpackage  PayPal - Sellacious Payments
 */
class plgSellaciousPaymentPayPal extends SellaciousPluginPayment
{
	/**
	 * Returns handlers to the payment methods that will be managed by this plugin
	 *
	 * @param   string  $context    The calling context, must be 'com_sellacious.payment' to effect
	 * @param   array   &$handlers  ByRef, associative array of handlers
	 *
	 * @return  bool
	 */
	public function onCollectHandlers($context, array &$handlers)
	{
		// PayPal library must exist, if not then skip
		if ($context == 'com_sellacious.payment' && class_exists('PayPal\Api\Payment'))
		{
			$handlers['paypal'] = JText::_('PLG_SELLACIOUSPAYMENT_PAYPAL_API_PAYPAL');
		}

		return true;
	}

	/**
	 * Triggers payment method to make a payment for the given transaction
	 * The order must have been created in the database prior calling this method.
	 * Since we need a redirection mechanism in between, we need to alter the calling sequence.
	 *
	 * @param   string  $context  The calling context, must be 'com_sellacious.payment' to effect
	 *
	 * @return  bool
	 * @throws  Exception
	 */
	public function onRequestPayment($context)
	{
		$app = JFactory::getApplication();

		$payment_id = $app->getUserState('com_sellacious.payment.execution.id');
		$handler    = $this->helper->payment->getFieldValue($payment_id, 'handler');

		$result   = true;
		$handlers = array();
		$this->onCollectHandlers($context, $handlers);

		if (array_key_exists($handler, $handlers))
		{
			$invoice = $this->getInvoice();
			$this->initPayment($invoice);
		}

		return $result;
	}

	/**
	 * Triggers payment execution on callback return after paypal approval
	 *
	 * @param   string  $context  The calling context, must be 'com_sellacious.payment' to effect
	 *
	 * @return  bool
	 * @throws  Exception
	 */
	public function onPaymentCallback($context)
	{
		$app = JFactory::getApplication();

		$payment_id = $app->getUserState('com_sellacious.payment.execution.id');
		$handler    = $this->helper->payment->getFieldValue($payment_id, 'handler');

		$result = true;

		if ($context == 'com_sellacious.payment' && $handler == 'paypal')
		{
			$response = $this->executePayment();
			$result   = $this->handleResponse($response);
		}

		return $result;
	}

	/**
	 * Sellacious do not bother about the details a plugin might need.
	 * Plugins are set free to fetch what they want. However basic data is directly accessible.
	 *
	 * @return  array  All the required details for the transaction execution with the Payment Gateway
	 * @throws  Exception
	 */
	protected function getInvoice()
	{
		$app = JFactory::getApplication();

		$payment_id = $app->getUserState('com_sellacious.payment.execution.id');
		$params     = $app->getUserState('com_sellacious.payment.execution.params', array());
		$payment    = $this->helper->payment->getItem($payment_id);

		$array = array(
			'params'      => $params,
			'amount'      => $payment->amount_payable,
			'currency'    => $payment->currency,
			'description' => $app->get('sitename'),
		);

		return $array;
	}

	/**
	 * Initialize SDK configurations using client key and secret with additional connection settings
	 *
	 * @throws  Exception
	 * @return  PayPal\Rest\ApiContext  PayPal SDK API context object
	 */
	protected function getApiContext()
	{
		$config = $this->getParams();
		$mode   = $config->get('api_mode', 'sandbox');

		if ($mode == 'sandbox')
		{
			$apiKey    = $config->get('apikey_sandbox');
			$apiSecret = $config->get('apisecret_sandbox');
		}
		else
		{
			$mode      = 'live';
			$apiKey    = $config->get('apikey');
			$apiSecret = $config->get('apisecret');
		}

		if (empty($apiKey) || empty($apiSecret))
		{
			throw new Exception(JText::_('PLG_SELLACIOUSPAYMENT_PAYPAL_API_CONFIG_NOT_SET'));
		}

		$api = array(
			'mode' => $mode,

			'http.ConnectionTimeOut' => $config->get('http_timeout', 30),
			'http.Retry'             => $config->get('http_retry', 1),
			'http.Proxy'             => $config->get('http_proxy', null),

			'log.LogEnabled' => $config->get('log_enabled', ''),
			'log.FileName'   => JPATH_SITE . '/logs/sellacious.payment.paypal.php',
			'log.LogLevel'   => $config->get('log_level', 'INFO'),
		);

		$credential = new PayPal\Auth\OAuthTokenCredential($apiKey, $apiSecret);
		$apiContext = new PayPal\Rest\ApiContext($credential, md5(time()));
		$apiContext->setConfig($api);

		return $apiContext;
	}

	/**
	 * Create a payment using the buyer's paypal account as the funding instrument.
	 * The app will have to redirect the buyer to the paypal website, obtain their consent to the payment
	 * and subsequently execute the payment using the execute API call.
	 *
	 * @param   mixed  $invoice  The data required by the payment gateway to execute the transaction
	 *
	 * @return  mixed  Transaction's response received from the gateway
	 * @throws  Exception
	 */
	protected function initPayment($invoice)
	{
		$returnUrl = $this->getCallbackUrl(array('status' => 'approved'));
		$cancelUrl = $this->getCallbackUrl(array('status' => 'cancelled'));

		$payer = new PayPal\Api\Payer();
		$payer->setPaymentMethod('paypal');

		$amount = new PayPal\Api\Amount();
		$amount->setCurrency($invoice['currency']);
		$amount->setTotal($invoice['amount']);

		$transaction = new PayPal\Api\Transaction();
		$transaction->setDescription($invoice['description']);
		$transaction->setAmount($amount);

		$redirects = new PayPal\Api\RedirectUrls();
		$redirects->setReturnUrl($returnUrl);
		$redirects->setCancelUrl($cancelUrl);

		$payment = new PayPal\Api\Payment();
		$payment->setIntent('sale');
		$payment->setPayer($payer);
		$payment->setRedirectUrls($redirects);
		$payment->setTransactions(array($transaction));

		$api = $this->getApiContext();
		$payment->create($api);

		if ($payment->getState() != 'created')
		{
			throw new Exception(JText::_('PLG_SELLACIOUSPAYMENT_PAYPAL_API_PAYPAL_PAYMENT_FAILED'));
		}

		// We need to redirect the user to PayPal for their approval of payment
		$app      = JFactory::getApplication();
		$redirect = $payment->getLink('approval_url');

		$app->redirect($redirect);
	}

	/**
	 * Execute the actual payment with the Payment Gateway.
	 *
	 * @return  PayPal\Api\Payment  Transaction's response received from the gateway
	 * @throws  Exception
	 */
	protected function executePayment()
	{
		$app     = JFactory::getApplication();
		$status  = $app->input->getString('status');
		$payment = null;

		if ($status == 'approved')
		{
			$paymentId = $app->input->get('paymentId', '', 'raw');
			$payerId   = $app->input->get('PayerID', '', 'raw');

			if (empty($paymentId) || empty($payerId))
			{
				throw new Exception(JText::_('PLG_SELLACIOUSPAYMENT_PAYPAL_INVALID_API_RESPONSE'));
			}

			$execution = new PayPal\Api\PaymentExecution();
			$execution->setPayerId($payerId);

			$api     = $this->getApiContext();
			$payment = new PayPal\Api\Payment();
			$payment->setId($paymentId);
			$payment->execute($execution, $api);
		}

		return $payment;
	}

	/**
	 * Generate response data to be stored in the transaction log and save
	 *
	 * @param   PayPal\Api\Payment  $payment  The Payment object as the PayPal response
	 *
	 * @return  bool
	 * @throws  Exception
	 */
	protected function handleResponse($payment)
	{
		$app = JFactory::getApplication();

		if ($payment instanceof PayPal\Api\Payment)
		{
			$transactions = $payment->getTransactions();
			$payerInfo    = $payment->getPayer()->getPayerInfo();

			$transaction = $transactions[0];
			$resources   = $transaction->getRelatedResources();

			$resource = $resources[0];
			$sale     = $resource->getSale();

			if ($payment->getState() != 'approved')
			{
				// failed, canceled, expired
				$status  = static::STATUS_DECLINED;
				$message = JText::_('PLG_SELLACIOUSPAYMENT_PAYPAL_API_PAYPAL_PAYMENT_FAILED');
			}
			elseif ($sale->getState() == 'completed')
			{
				$status  = static::STATUS_APPROVED;
				$message = JText::_('PLG_SELLACIOUSPAYMENT_PAYPAL_API_PAYPAL_PAYMENT_COMPLETE');
			}
			elseif ($sale->getState() == 'pending')
			{
				if ($sale->getReasonCode() == 'PAYMENT_REVIEW')
				{
					$status  = static::STATUS_APPROVAL_HOLD;
					$message = JText::_('PLG_SELLACIOUSPAYMENT_PAYPAL_API_PAYPAL_PAYMENT_PENDING_REVIEW');
				}
				else
				{
					$status  = static::STATUS_APPROVAL_HOLD;
					$message = JText::_('PLG_SELLACIOUSPAYMENT_PAYPAL_API_PAYPAL_PAYMENT_PENDING');
				}
			}
			else
			{
				// Refunded, partially refunded. Does not apply in our case probably!
				$status  = static::STATUS_DECLINED;
				$message = JText::_('PLG_SELLACIOUSPAYMENT_PAYPAL_API_PAYPAL_PAYMENT_FAILED');
			}

			$registry = new Joomla\Registry\Registry;
			$registry->set('transaction_id', $payment->getId());
			$registry->set('payer_id', $payerInfo->getPayerId());
			$registry->set('sale_id', $sale->getId());
			$registry->set('payer_info', $payerInfo->toArray());
			$registry->set('created', $payment->getCreateTime());
			$registry->set('updated', $payment->getUpdateTime());
			$registry->set('reason_code', $sale->getReasonCode());
			$registry->set('payment_state', $payment->getState());
			$registry->set('sale_state', $sale->getState());
			$registry->set('amount', $transaction->getAmount()->getTotal());
			$registry->set('currency', $transaction->getAmount()->getCurrency());

			$data = array(
				'response_code'    => 'PAYPAL:' . $payment->getState(),
				'response_state'   => $payment->getState(),
				'response_message' => $message,
				'response_data'    => $registry->toArray(),
				'transaction_id'   => $payment->getId(),
				'state'            => $status,
			);
		}
		else
		{
			$status  = static::STATUS_ABORTED;
			$token   = $app->input->get('token', '', 'raw');
			$message = JText::_('PLG_SELLACIOUSPAYMENT_PAYPAL_USER_REJECTED_TRANSACTION');

			$data = array(
				'response_code'    => 'PAYPAL:aborted',
				'response_state'   => 'aborted',
				'response_message' => $message,
				'response_data'    => json_encode(array('token' => $token)),
				'transaction_id'   => '',
				'state'            => static::STATUS_ABORTED,
			);
		}

		$config = $this->getParams();
		$mode   = $config->get('api_mode', 'sandbox');
		$data   = ArrayHelper::toObject($data);

		$this->saveResponse($data->response_code, $data->response_state, $data->response_message, $data->response_data, $data->transaction_id, $data->state, $mode == 'sandbox');

		return $status == 1;
	}
}
